/*
**
**  WAY (Where Are You ?)
**
**  info.c		additional information handling module.
**
**  Author: Roberto Bagnara (bagnara@vxcern.cern.ch),
**          CERN  DD Division OC Group RA Section.
**
**  All rights reserved to:
**
**          CERN
**          European Laboratory for Particle Physics
**          CH 1211 Geneva 23
**          SWITZERLAND
**
**  Permission is granted to any individual or institution
**  to use, copy, or redistribute this software so long as
**  it  is not  sold for  profit, provided  this copyright
**  notice is retained.
**
**  Modification History:
**
**  Version  Date    Who              Comments
**
**  1.0.00   890901  Roberto Bagnara  First working version
**
**
**  Description:
**
**  this module contains the procedures for handling the additional
**  informations that the name server keeps for any regitration.
**  As it is not well know yet which kind of informations will be
**  necessary this separate module allows for a useful bit of abstraction
**  and independence.
**
*/

/*
**      Include files.
**      --------------
*/

#include "way.h"
#include <strings.h>
#include "info.h"

/*
**      External functions.
**      -------------------
*/

#ifdef __STDC__
extern int atoi(const char *nptr);
#else
extern int atoi();
#endif

/*
**      External functions that don't like too much checking.
**      -----------------------------------------------------
*/

extern void fatal(/* function, format, arg1, arg2, ... */);

/*
**	This is the keyword table.
**	--------------------------
**
**	This table contains each info keyword with its type and an additional
**	"magic" field. When the keyword type is FLAG "magic" provides the
**	indication of which bits should be set or cleared. If "magic" is
**	positive, its binary representation corresponds to the bits to set.
**	If "magic" is negative, the binary representation of its 2's
**	complement corresponds to the bits to clear. If the keyword type
**	is NUMERIC the "magic" field contains the index to access the
**	correspondent long variable in the info structure. The STRING keyword
**	type is not implemented yet.
**
**	The table is initialized from a macro defined in info.h to
**	where you are referred for more info.
*/

struct {
    int  type;
    char *name;
    long magic;
} info_keywords[] = INFO_KEYWORDS;

/*
**
**	Return the type of a keyword.
**	-----------------------------
**
**
**	int keyword_type(const char *keyword)
**
**  Parameters:	keyword		keyword to check.
**
**  Return:			keyword type or AMBIGUOUS or UNKNOWN.
**
**  On entry:	none
**
**  On exit:	none
**
**  Description:
**
**	Scan the keyword table trying to find an entry that unambiguously
**	matches the given (possibly truncated) keyword. If such a match
**	exists then the keyword type is returned (FLAG, NUMERIC, STRING etc.).
**	If the given keyword is ambiguous (there are more than one match) or
**	unknown (no match at all) the correspondent indication is returned.
**	This procedure is used by the parser routine to know how to parse the
**	text following a keyword.
**
**  History:
**
**	01 Sep 89	Written. (RB)
**
*/

#ifdef __STDC__
int keyword_type(const char *keyword)
#else
int keyword_type(keyword)
char *keyword;
#endif
{
    register int i, j, k, l;

    /* j is initialized here only to avoid a warning. */
    
    for (i = j = k = 0, l = strlen(keyword);
	  k < 2 && i < sizeof(info_keywords) / sizeof(info_keywords[0]); i++)
	if (!strncmp(info_keywords[i].name, keyword, l)) {
	    j = i;
	    k++;
	}
    if (k == 0)
	return UNKNOWN;
    else if (k > 1)
	return AMBIGUOUS;
    else
	return info_keywords[j].type;
}

/*
**
**	Process a keyword.
**	------------------
**
**
**	void process_keyword(const char *keyword, const char *value, INFO_PTR info)
**
**  Parameters:	keyword		keyword to process;
**		value		value associated to the keyword, if any;
**		info		info structure to be modified.
**
**  Return:	none
**
**  On entry:	none
**
**  On exit:	none
**
**  Description:
**
**	Scan the keyword table trying to find an entry that unambiguously
**	matches the given (possibly truncated) keyword. If such a match
**	exists then the keyword is processed according to its type.
**	If the keyword is a FLAG then the associated bit in the "info"
**	structure is set/reset accordingly. If the keyword is NUMERIC then
**	the correspondent variable in the "info" structure is set to the
**	number whose representation is "value". STRING keywords are not
**	yet. If "keyword" is ambiguous or unknown a fatal error occurs.
**
**  History:
**
**	01 Sep 89	Written. (RB)
**
*/

#ifdef __STDC__
void process_keyword(const char *keyword, const char *value, INFO_PTR info)
#else
void process_keyword(keyword, value, info)
char *keyword;
char *value;
INFO_PTR info;
#endif
{
    register int i, j, k, l;

    /* j is initialized here only to avoid a warning. */

    for (i = j = k = 0, l = strlen(keyword);
	  k < 2 && i < sizeof(info_keywords) / sizeof(info_keywords[0]); i++)
	if (!strncmp(info_keywords[i].name, keyword, l)) {
	    j = i;
	    k++;
	}
    if (k == 0)
      fatal("process_keyword", "invalid keyword specification `%s'", keyword);
    else if (k > 1)
      fatal("process_keyword", "ambiguous keyword specification `%s'", keyword);
    else if (info_keywords[j].type == FLAG)
	if (info_keywords[j].magic < 0)
	    info->flags &= ~-info_keywords[j].magic;
	else
	    info->flags |= info_keywords[j].magic;
    else if (info_keywords[j].type == NUMERIC)
	info->long_info.array[info_keywords[j].magic] = atoi(value);
}
