/*
**
**  WAY (Where Are You ?)
**
**  interfc.c		interface module.
**
**  Author: Roberto Bagnara (bagnara@vxcern.cern.ch),
**          CERN  DD Division OC Group RA Section.
**
**  All rights reserved to:
**
**          CERN
**          European Laboratory for Particle Physics
**          CH 1211 Geneva 23
**          SWITZERLAND
**
**  Permission is granted to any individual or institution
**  to use, copy, or redistribute this software so long as
**  it  is not  sold for  profit, provided  this copyright
**  notice is retained.
**
**  Modification History:
**
**  Version  Date    Who              Comments
**
**  1.0.00   891030  Roberto Bagnara  First working version
**
**
**  Description:
**
**  This module contains the interface procedures used to get access to the
**  WAY services from the outside world. In addition, a number of procedures
**  allowing WAY remote control are provided here.
**
*/

/*
**	Include files.
**	--------------
*/

#include <strings.h>
#include "way.h"
#include "info.h"
#include "dbase.h"

/*
**	External functions.
**	-------------------
*/

#ifdef __STDC__
extern void db_associate(const char *service, const char *address, const INFO info);
extern void db_dissociate(const char *service, const char *address);
extern LIST_PTR select_best_address(const char *service, int version,
			int compatibility, const char *pref_addresses);
#else
extern void db_associate();
extern void db_dissociate();
extern LIST_PTR select_best_address();
#endif

/*
**      External functions that don't like too much checking.
**      -----------------------------------------------------
*/

extern void wlog(/* format, arg1, arg2, ... */);
extern void tlog(/* format, arg1, arg2, ... */);

/*
**	External variables.
**	-------------------
*/

extern int log_level;

/*
**	The following code is provisional !!!
**	-------------------------------------
*/

#ifdef __STDC__
way_status way_register(const char *service,
			int version, int compatibility, const char *address)
#else
way_status way_register(service, version, compatibility, address)
char *service;
int version;
int compatibility;
char *address;
#endif
{
    INFO info;

    /* Record version and compatibility into the info structure. */

    info.long_info.longs.version = version;
    info.long_info.longs.compatibility = compatibility;

    /* Leave a track of what we are doing. */

    LOG(1) tlog("Registering `%s' (Version %d, Backward %d) at `%s'.",
	service, version, compatibility, address);

    /* Insert into the database. */

    db_associate(service, address, info);

    return WAY_S_NORMAL;
}

#ifdef __STDC__
way_status way_unregister(const char *service, int version, const char *address)
#else
way_status way_unregister(service, version, address)
char *service;
int version;
char *address;
#endif
{
    /* Leave a track of what we are doing. */

    LOG(1) tlog("Unregistering `%s' (Version %d) at `%s' .",
	service, version, address);

    /* Remove from the database. */

    db_dissociate(service, address);

    return WAY_S_NORMAL;
}

#ifdef __STDC__
way_status way_request(const char *service, int version, int compatibility,
	const char *pref_addresses, int  address_length, char *address)
#else
way_status way_request(service, version, compatibility,
				     pref_addresses, address_length, address)
char *service;
int version;
int compatibility;
char *pref_addresses;
int  address_length;
char *address;
#endif
{
    LIST_PTR tmp;

    LOG(1) {
	tlog("Request for `%s' (Version %d, Backward %d) from `%s'.",
	    service, version, compatibility, "(unknown)");
	wlog("Allowed addresses are `%s'.", pref_addresses);
    }

    /* See if an appropriate address can be found in the database. */

    tmp = select_best_address(service, version, compatibility, pref_addresses);

    /* If not just report this to the caller. */

    if (tmp == NULL_LIST_PTR) {
	LOG(1) wlog("No suitable server is known.");
	return WAY_S_NO_SERVER;
    }

    /* If there is one we must ensure we can report it correctly. */

    if (strlen(tmp->a_node->name) > address_length - 1) {
	LOG(1) wlog("Client-supplied buffer too short (%d) to hold `%s'.",
	    address_length, tmp->a_node->name);
	return WAY_S_BUFFER_OVERFLOW;
    }

    (void) strcpy(address, tmp->a_node->name);
    LOG(1) wlog("`%s' is a suitable server: RETURNED.", tmp->a_node->name);
    return WAY_S_NORMAL;
}
