/*
**
**  WAY (Where Are You ?)
**
**  remote.c		remote control module.
**
**  Author: Roberto Bagnara (bagnara@vxcern.cern.ch),
**          CERN  DD Division OC Group RA Section.
**
**  All rights reserved to:
**
**          CERN
**          European Laboratory for Particle Physics
**          CH 1211 Geneva 23
**          SWITZERLAND
**
**  Permission is granted to any individual or institution
**  to use, copy, or redistribute this software so long as
**  it  is not  sold for  profit, provided  this copyright
**  notice is retained.
**
**  Modification History:
**
**  Version  Date    Who              Comments
**
**  1.0.00   891102  Roberto Bagnara  First working version
**
**
**  Description:
**
**  this module contains the procedures used for WAY remote control.
**
*/

/*
**	Include files.
**	--------------
*/

#include "way.h"
#include <strings.h>

/*
**	External functions.
**	-------------------
*/

#ifdef __STDC__
extern void exit(int status);
#else
extern void exit();
#endif

/*
**      External functions that don't like too much checking.
**      -----------------------------------------------------
*/

extern void wlog(/* format, arg1, arg2, ... */);
extern void tlog(/* format, arg1, arg2, ... */);

/*
**	External variables.
**	-------------------
*/

extern int log_level;
extern char password[MAX_PWD_LENGTH];

/*
way_status way_rc_fload(char *file_name)
way_status way_rc_fdump(char *file_name)
way_status way_rc_change(...)
*/

/*
**
**	WAY shutdown.
**	-------------
**
**
**	way_status way_rc_shutdown(const char *passwd)
**
**  Parameters:	passwd		authorization password.
**
**  Return:			WAY_S_NORMAL on success (see below);
**				WAY_S_NOPRIV on failure.
**
**  On entry:	none
**
**  On exit:	none
**
**  Description:
**
**	Check the client-provided authorization password "passwd" for
**	validity. If the check is successful then stop all the WAY
**	activities and quit. If the check fails then WAY_S_NOPRIV is
**	returned.
**
**				NOTE
**
**	This very preliminary version of the routine just calls exit(),
**	and is thus to be declared, in the RPCL specification file, with
**	PRAGMA CONCURRENT.
**
**	This means, in particular that IF THE PASSWORD CHECK IS SUCCESSFUL
**	THIS ROUTINE WONT RETURN.
**
**	Future revisions, depending also on the type of server chosen
**	(i.e. simple loop, polling, asynchronous), may (and very likely
**	will) use a different strategy for stopping.
**
**  History:
**
**	11 Nov 89	Written. (RB)
**
*/

#ifdef __STDC__
way_status way_rc_shutdown(const char *passwd)
#else
way_status way_rc_shutdown(passwd)
char *passwd;
#endif
{
    if (!strcmp(passwd, password)) {
    	LOG(1) tlog("Shutdown requested by operator %s: GOING DOWN.",
    	    "(unknown)");
	exit(SUCCESS_EXIT_CODE);
	return WAY_S_NORMAL;
    } else {
	LOG(1) tlog("Shutdown requested by unauthorized operator %s: REFUSED.",
	    "(unknown)");
    	return WAY_S_NOPRIV;
    }
}

/*
**
**	Change authorization password.
**	------------------------------
**
**
**	way_status way_rc_change_password(const char *oldpwd,
**					  const char *newpwd)
**
**  Parameters:	oldpwd		old password;
**				new password.
**
**  Return:			WAY_S_NORMAL on success;
**				WAY_S_NOPRIV on failure.
**
**  On entry:	none
**
**  On exit:	none
**
**  Description:
**
**	Check the client-provided authorization password "oldpwd" for
**	validity. If the check is successful then set the new authorization
**	password to "newpwd" and return WAY_S_NORMAL. If the check fails
**	then WAY_S_NOPRIV is returned.
**
**  History:
**
**	11 Nov 89	Written. (RB)
**
*/

#ifdef __STDC__
way_status way_rc_change_password(const char *oldpwd, const char *newpwd)
#else
way_status way_rc_change_password(oldpwd, newpwd)
char *oldpwd;
char *newpwd;
#endif
{
    if (!strcmp(oldpwd, password)) {
    	LOG(1) tlog("Password change requested by operator %s: CHANGED.",
	    "(unknown)");
    	(void) strcpy(password, newpwd);
	return WAY_S_NORMAL;
    } else {
    	LOG(1) tlog("Password change requested by unauthorized operator %s: REFUSED.",
	    "(unknown)");
    	return WAY_S_NOPRIV;
    }
}
